@@ -19,6 +19,7 @@ use crate::config as cfg;
 use lazy_static::lazy_static;
 use std::collections::HashMap;
 use std::io::{BufRead, Write};
+use std::path as p;
 use std::process as proc;
 
 pub fn get_swayr_socket_path() -> String {
@@ -43,25 +44,44 @@ pub fn get_swayr_socket_path() -> String {
     )
 }
 
-fn desktop_entries() -> Vec<String> {
-    let mut dirs = vec![];
-    if let Some(dd) = directories::BaseDirs::new()
-        .map(|b| b.data_local_dir().to_string_lossy().to_string())
-    {
-        dirs.push(dd);
-    }
-    dirs.push(String::from("/usr/share/applications/"));
+fn desktop_entry_folders() -> Vec<Box<p::Path>> {
+    let mut dirs: Vec<Box<p::Path>> = vec![];
 
+    // XDG_DATA_HOME/applications
+    if let Some(dd) = directories::BaseDirs::new() {
+        dirs.push(dd.data_local_dir().to_path_buf().into_boxed_path());
+    }
+
+    let default_dirs =
+        ["/usr/local/share/applications/", "/usr/share/applications/"];
+    for dir in default_dirs {
+        dirs.push(p::Path::new(dir).to_path_buf().into_boxed_path());
+    }
+
+    if let Ok(xdg_data_dirs) = std::env::var("XDG_DATA_DIRS") {
+        for mut dir in std::env::split_paths(&xdg_data_dirs) {
+            dir.push("applications/");
+            dirs.push(dir.into_boxed_path());
+        }
+    }
+
+    dirs.sort();
+    dirs.dedup();
+
+    dirs
+}
+
+fn desktop_entries() -> Vec<Box<p::Path>> {
     let mut entries = vec![];
-    for dir in dirs {
-        if let Ok(readdir) = std::fs::read_dir(dir) {
+    for dir in desktop_entry_folders() {
+        if let Ok(readdir) = dir.read_dir() {
             for entry in readdir.flatten() {
                 let path = entry.path();
                 if path.is_file()
                     && path.extension().map(|ext| ext == "desktop")
                         == Some(true)
                 {
-                    entries.push(path.as_path().to_string_lossy().to_string());
+                    entries.push(path.to_path_buf().into_boxed_path());
                 }
             }
         }
@@ -69,39 +89,50 @@ fn desktop_entries() -> Vec<String> {
     entries
 }
 
-fn find_icon(icon_name: &str, icon_dirs: &[String]) -> Option<String> {
-    if std::path::Path::new(icon_name).is_file() {
-        return Some(String::from(icon_name));
+fn find_icon(icon_name: &str, icon_dirs: &[String]) -> Option<Box<p::Path>> {
+    let p = p::Path::new(icon_name);
+    if p.is_file() {
+        println!("(1) Icon name '{}' -> {}", icon_name, p.display());
+        return Some(p.to_path_buf().into_boxed_path());
     }
 
     for dir in icon_dirs {
         for ext in &["png", "svg"] {
-            let mut pb = std::path::PathBuf::from(dir);
-            pb.push(icon_name);
-            pb.set_extension(ext);
+            let mut pb = p::PathBuf::from(dir);
+            pb.push(icon_name.to_owned() + "." + ext);
             let icon_file = pb.as_path();
             if icon_file.is_file() {
-                return Some(String::from(icon_file.to_str().unwrap()));
+                println!(
+                    "(2) Icon name '{}' -> {}",
+                    icon_name,
+                    icon_file.display()
+                );
+                return Some(icon_file.to_path_buf().into_boxed_path());
             }
         }
     }
 
+    println!("(3) No icon for name {}", icon_name);
     None
 }
 
 lazy_static! {
     static ref WM_CLASS_OR_ICON_RX: regex::Regex =
         regex::Regex::new("(StartupWMClass|Icon)=(.+)").unwrap();
+    static ref REV_DOMAIN_NAME_RX: regex::Regex =
+        regex::Regex::new(r"^(?:[a-zA-Z0-9-]+\.)+([a-zA-Z0-9-]+)$").unwrap();
 }
 
-fn get_app_id_to_icon_map(icon_dirs: &[String]) -> HashMap<String, String> {
-    let mut map = HashMap::new();
+fn get_app_id_to_icon_map(
+    icon_dirs: &[String],
+) -> HashMap<String, Box<p::Path>> {
+    let mut map: HashMap<String, Box<p::Path>> = HashMap::new();
 
     for e in desktop_entries() {
         if let Ok(f) = std::fs::File::open(&e) {
             let buf = std::io::BufReader::new(f);
             let mut wm_class: Option<String> = None;
-            let mut icon: Option<String> = None;
+            let mut icon: Option<Box<p::Path>> = None;
 
             // Get App-Id and Icon from desktop file.
             for line in buf.lines() {
@@ -124,45 +155,63 @@ fn get_app_id_to_icon_map(icon_dirs: &[String]) -> HashMap<String, String> {
             }
 
             if let Some(icon) = icon {
+                // Sometimes the StartupWMClass is the app_id, e.g. FF Dev
+                // Edition has StartupWMClass firefoxdeveloperedition although
+                // the desktop file is named firefox-developer-edition.
                 if let Some(wm_class) = wm_class {
                     map.insert(wm_class, icon.clone());
                 }
-                map.insert(
-                    String::from(
-                        std::path::Path::new(&e)
-                            .with_extension("")
-                            .file_name()
-                            .unwrap()
-                            .to_string_lossy(),
-                    ),
-                    icon,
+
+                // Some apps have a reverse domain name desktop file, e.g.,
+                // org.gnome.eog.desktop but reports as just eog.
+                let desktop_file_name = String::from(
+                    e.with_extension("").file_name().unwrap().to_string_lossy(),
                 );
+                if let Some(caps) =
+                    REV_DOMAIN_NAME_RX.captures(&desktop_file_name)
+                {
+                    map.insert(
+                        caps.get(1).unwrap().as_str().to_string(),
+                        icon.clone(),
+                    );
+                }
+
+                // The usual case is that the app with foo.desktop also has the
+                // app_id foo.
+                map.insert(desktop_file_name.clone(), icon);
             }
         }
     }
 
+    println!(
+        "Desktop entries to icon files ({} entries):\n{:#?}",
+        map.len(),
+        map
+    );
     map
 }
 
 lazy_static! {
-    static ref APP_ID_TO_ICON_MAP: std::sync::Mutex<Option<HashMap<String, String>>> =
+    static ref APP_ID_TO_ICON_MAP: std::sync::Mutex<Option<HashMap<String, Box<p::Path>>>> =
         std::sync::Mutex::new(None);
 }
 
-pub fn get_icon(app_id: &str, icon_dirs: &[String]) -> Option<String> {
+pub fn get_icon(app_id: &str, icon_dirs: &[String]) -> Option<Box<p::Path>> {
     let mut opt = APP_ID_TO_ICON_MAP.lock().unwrap();
 
     if opt.is_none() {
         opt.replace(get_app_id_to_icon_map(icon_dirs));
     }
 
-    opt.as_ref().unwrap().get(app_id).map(String::from)
+    opt.as_ref().unwrap().get(app_id).map(|i| i.to_owned())
 }
 
 #[test]
-fn test_desktop_entries() {
+fn test_icon_stuff() {
     let icon_dirs = vec![
+        String::from("/usr/share/icons/hicolor/scalable/apps"),
         String::from("/usr/share/icons/hicolor/48x48/apps"),
+        String::from("/usr/share/icons/Adwaita/48x48/apps"),
         String::from("/usr/share/pixmaps"),
     ];
     let m = get_app_id_to_icon_map(&icon_dirs);
@@ -194,27 +243,19 @@ where
         map.insert(s, c);
     }
 
-    let menu_default = cfg::Menu::default();
-    let menu_exec = cfg
-        .menu
-        .as_ref()
-        .and_then(|l| l.executable.as_ref())
-        .unwrap_or_else(|| menu_default.executable.as_ref().unwrap());
+    let menu_exec = cfg.get_menu_executable();
     let args: Vec<String> = cfg
-        .menu
-        .as_ref()
-        .and_then(|l| l.args.as_ref())
-        .unwrap_or_else(|| menu_default.args.as_ref().unwrap())
+        .get_menu_args()
         .iter()
         .map(|a| a.replace("{prompt}", prompt))
         .collect();
 
-    let mut menu = proc::Command::new(menu_exec)
+    let mut menu = proc::Command::new(&menu_exec)
         .args(args)
         .stdin(proc::Stdio::piped())
         .stdout(proc::Stdio::piped())
         .spawn()
-        .expect(&("Error running ".to_owned() + menu_exec));
+        .expect(&("Error running ".to_owned() + &menu_exec));
 
     {
         let stdin = menu
